#!/usr/bin/env python
#
# bhv - bhv.py
# Author: Loren Bandiera
# Copyright 2005 MMG Security, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
#

import sys
import time
import csv
import pygtk
import gtk
import gtk.glade
import gnome
import gnome.ui
import ie
import demork
import defines

class bhv:
	def __init__ (self):
		self.gnome_program = gnome.init(defines.PACKAGE, defines.VERSION)
		self.ie = None
		self.db = None
		self.current_file_type = 0

		gladefile = "%s/browser-history-viewer.glade" % defines.PROG_DIR
		self.Glade = gtk.glade.XML (gladefile)

		# setup main window
		self.mainwindow = self.Glade.get_widget ("mainWindow")
		self.mainwindow.resize (640, 480)

		# get/hide scrolled windows for now
		self.ie_scrolled_window = self.Glade.get_widget ("scrolledwindow7")
		self.ie_scrolled_window.hide ()

		self.ff_scrolled_window = self.Glade.get_widget ("scrolledwindow8")
		self.ff_scrolled_window.hide ()

		# setup ie treeview
		self.ietreeview = self.Glade.get_widget ("ieTreeView")
		self.iestore = gtk.ListStore (str, str, str, str)
		self.ietreeview.set_model (self.iestore)

		c0 = gtk.TreeViewColumn("Access time", gtk.CellRendererText(), text=0)
		c0.set_resizable (True)
		c0.set_spacing (5)
		c0.set_sort_column_id (0)

		c1 = gtk.TreeViewColumn("Modify time", gtk.CellRendererText(), text=1)
		c1.set_resizable (True)
		c1.set_spacing (5)
		c1.set_sort_column_id (1)

		c2 = gtk.TreeViewColumn("Url", gtk.CellRendererText(), text=2)
		c2.set_resizable (True)
		c2.set_spacing (5)
		c2.set_sort_column_id (2)

		c3 = gtk.TreeViewColumn("Filename", gtk.CellRendererText(), text=3)
		c3.set_resizable (True)
		c3.set_spacing (5)
		c3.set_sort_column_id (3)
	
		self.ietreeview.append_column (c0)
		self.ietreeview.append_column (c1)
		self.ietreeview.append_column (c2)
		self.ietreeview.append_column (c3)
		self.ietreeview.connect ("button-press-event", self.on_right_click);

		# setup ff treeview
		self.fftreeview = self.Glade.get_widget ("ffTreeView")
		self.ffstore = gtk.ListStore (str, str, str, str)
		self.fftreeview.set_model (self.ffstore)

		c4 = gtk.TreeViewColumn("First access time", gtk.CellRendererText(), text=0)
		c4.set_resizable (True)
		c4.set_spacing (5)
		c4.set_sort_column_id (0)

		c5 = gtk.TreeViewColumn("Last access time", gtk.CellRendererText(), text=1)
		c5.set_resizable (True)
		c5.set_spacing (5)
		c5.set_sort_column_id (1)

		c6 = gtk.TreeViewColumn("Times accessed", gtk.CellRendererText(), text=2)
		c6.set_resizable (True)
		c6.set_spacing (5)
		c6.set_sort_column_id (2)

		c7 = gtk.TreeViewColumn("URL", gtk.CellRendererText(), text=3)
		c7.set_resizable (True)
		c7.set_spacing (5)
		c7.set_sort_column_id (3)
	
		self.fftreeview.append_column (c4)
		self.fftreeview.append_column (c5)
		self.fftreeview.append_column (c6)
		self.fftreeview.append_column (c7)
		self.fftreeview.connect ("button-press-event", self.on_right_click);

		# connect event handlers
		dic = { "OnOpenActivate" : self.on_open_activate,
			"OnSaveActivate" : self.on_save_activate,
			"OnQuitActivate" : (gtk.main_quit),
			"OnAppDelete" : (gtk.main_quit),
			"OnAboutActivate" : self.on_about_activated,
			"OnHelpContentsActivate" : self.on_help_activated }
		
		self.Glade.signal_autoconnect (dic)
		return

	def check_file_type (self, filename):
		f = open (filename, "r")
		line = f.readline ()
		f.seek (0x18)
		version = f.read (4)
		f.close ()

		if (line.find ('<mdb:mork') >= 0):
			return 1;
		elif (version == 4 or version >= 5):
			return 2;

		return 0;
	
	def handle_mork (self, db):
		columns = db.cdict.keys()
		columns.sort(demork.hexcmp)

		tables = db.tables.keys()
		tables.sort(demork.hexcmp)

		for table in [ db.tables[k] for k in tables ]:
			rows = table.rows.keys()
			rows.sort(demork.hexcmp)
					
			url = ""
			firsttime = ""
			lasttime = ""
			visitcount = ""

			for row in [ table.rows[k] for k in rows ]:
				for cell in row.cells:
					if (cell.column == 'URL'):
						url = cell.atom
					elif (cell.column == 'FirstVisitDate'):
						tmp = int(cell.atom) / 1000000
						firsttime = time.ctime (tmp)
					elif (cell.column == 'LastVisitDate'):
						tmp = int(cell.atom) / 1000000
						lasttime = time.ctime (tmp)
					elif (cell.column == 'VisitCount'):
						visitcount = cell.atom

				self.ffstore.append ([firsttime, lasttime, visitcount, url])

	def on_open_activate (self, widget):
		dialog = gtk.FileChooserDialog("Select history file to examine",
		                               None,
		                               gtk.FILE_CHOOSER_ACTION_OPEN,
		                               (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
		                                gtk.STOCK_OPEN, gtk.RESPONSE_OK))

		dialog.set_default_response(gtk.RESPONSE_OK)

		filter = gtk.FileFilter()
		filter.set_name("Internet Explorer")
		filter.add_pattern("*.dat")
		dialog.add_filter(filter)

		filter = gtk.FileFilter()
		filter.set_name("Mozilla/Firefox")
		filter.add_pattern("*.dat")
		dialog.add_filter(filter)

		filter = gtk.FileFilter()
		filter.set_name("All files")
		filter.add_pattern("*")
		dialog.add_filter(filter)

		response = dialog.run()
		if response == gtk.RESPONSE_OK:
			filename = dialog.get_filename()
			self.current_file_type = self.check_file_type (filename)

			if (self.current_file_type == 1):
				self.iestore.clear ()
				self.ie_scrolled_window.hide ()
				self.ff_scrolled_window.show ()
				self.ffstore.clear ()

				self.db = demork.parse_mork (filename)
				self.handle_mork (self.db)

			elif (self.current_file_type == 2):
				self.ffstore.clear ()
				self.ff_scrolled_window.hide ()
				self.ie_scrolled_window.show ()
				self.iestore.clear ()

				self.ie = ie.IE (filename)
				self.ie.parse ()
			
				for h in self.ie.browserHistory[:]:
					self.iestore.append ([h.atime, h.mtime, h.url, h.filename])


		elif response == gtk.RESPONSE_CANCEL:
		    print 'Closed, no files selected'
		dialog.destroy()
	
	def save_csv_history (self, filename):
		writer = csv.writer (open(filename, "wb"))

		for h in self.ie.browserHistory[:]:
			row = [h.atime, h.mtime, h.url, h.filename]
			writer.writerow (row)

	def save_tab_history (self, filename):
		f = open(filename, "wb")

		for h in self.ie.browserHistory[:]:
			line = "%s\t%s\t%s\t%s\n" % (h.atime, h.mtime, h.url, h.filename)
			f.write (line)

		f.close ()


	def on_save_activate (self, widget):
		dialog = gtk.FileChooserDialog("Select data as...",
		                               None,
		                               gtk.FILE_CHOOSER_ACTION_SAVE,
		                               (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
		                                gtk.STOCK_SAVE, gtk.RESPONSE_OK))

		dialog.set_default_response(gtk.RESPONSE_OK)

		if (self.current_file_type == 1):  
			filter = gtk.FileFilter()
			filter.set_name("Extensible Markup Language (XML)")
			filter.add_pattern("*.xml")
			dialog.add_filter(filter)

		elif (self.current_file_type == 2):
			filter = gtk.FileFilter()
			filter.set_name("Comma Seperated Value (CSV)")
			filter.add_pattern("*.csv")
			dialog.add_filter(filter)

			filter = gtk.FileFilter()
			filter.set_name("Tab Delimited")
			filter.add_pattern("*.txt")
			dialog.add_filter(filter)

		response = dialog.run()
		if response == gtk.RESPONSE_OK:
			nf = dialog.get_filter().get_name()

			if (nf == "Comma Seperated Value (CSV)"):
				self.save_csv_history (dialog.get_filename())
			elif (nf == "Tab Delimited"):
				self.save_tab_history (dialog.get_filename())
			elif (nf == "Extensible Markup Language (XML)"):
				demork.outputMindy (self.db, dialog.get_filename())

		elif response == gtk.RESPONSE_CANCEL:
		    print 'Closed, no files selected'
		dialog.destroy()

	def on_open_url (self, widget):
		url = ""

		if (self.current_file_type == 1):
			model, row = self.fftreeview.get_selection().get_selected()
			url = self.ffstore.get_value (row, 3)
		else:
			model, row = self.ietreeview.get_selection().get_selected()
			url = self.iestore.get_value (row, 2)

		gnome.url_show (url)

	def on_right_click (self, widget, event):
		if event.button == 3:
			popup = gtk.Menu ()

			item = gtk.MenuItem ("Open URL")
			item.connect ("activate", self.on_open_url)
			item.show ()

			popup.append (item)
			popup.popup (None, None, None, 3, event.time)

	def on_help_activated (self, widget):
		gnome.help_display_desktop(self.gnome_program, "bhv", "bhv.xml", "")

	def on_about_activated (self, widget):
		a = gtk.AboutDialog ()
		a.set_name (defines.PACKAGE)
		a.set_version (defines.VERSION)
		a.set_copyright ("Copyright (c) MMG Security, Inc.")
		a.set_comments ("Browser history viewer")
		a.set_authors (['Loren Bandiera'])
		a.run ()
		a.destroy ()

################################################
# main
################################################
def main ():
	app = bhv ()
	gtk.main ()


result = main ()
sys.exit (result)
