#!/usr/bin/env python
#
# bhv - ie.py
# Author: Loren Bandiera
# Copyright 2005 MMG Security, Inc.
#
# Code ported from pasco.c 
# Copyright (C) 2003, by Keith J. Jones.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of the project nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

import sys

################################################
# History
################################################
class History:
	def __init__ (self):
		self.url = ""
		self.dirname = ""
		self.filename = ""
		self.mtime = ""
		self.atime = ""
		self.headers = ""


################################################
# IE
################################################
class IE:
	def __init__ (self, filename):
		self.browserHistory = [ ]
		self.filename = filename

	def bah_to_i (self, value):
		size = len (value)
		total = 0
		i = 0;

		for c in value:
			total += (ord(c) << 8 * i)
			i += 1
		
		return total

	def win_time_to_unix (self, value):
		fourbytes = value[0:3];
		low = self.bah_to_i (fourbytes)
		
		fourbytes = value[4:7]
		high = self.bah_to_i (fourbytes)
		
		dbl = high * pow (2, 32)
		dbl += low
		dbl *= 1.0e-7;
		dbl -= 11644473600;

		return dbl
		
	def parseRedr (self, f, currecoff, filesize, rtype):
		f.seek (currecoff + 0x10)
		c = f.read (1)
		i = 0
		url = ""

		while (c[0] != '\0' and currecoff + 0x10 + i + 1 < filesize):
			url += c[0]
			f.seek (currecoff + 0x10 + i + 1)
			c = f.read (1)
			i += 1

		h = History ()
		h.url = url

		self.browserHistory.append (h)


	def parseUrl (self, f, currecoff, filesize, rtype, fileversion):
		f.seek (currecoff + 4)
		tmp = f.read (4)
		reclen = self.bah_to_i (tmp)
		reclen *= 0x80

		h = History ()

		f.seek (currecoff + 8)
		mtime = f.read (8)
		h.mtime = self.win_time_to_unix (mtime)

		f.seek (currecoff + 8)
		atime = f.read (8)
		h.atime = self.win_time_to_unix (atime)

		newoffset = 0

		if (fileversion >= 5):
			newoffset = currecoff + 0x34
		else:
			newoffset = currecoff + 0x38

		f.seek (newoffset)
		tmp = f.read (1)
		urloff = int (ord(tmp[0]))

		f.seek (currecoff + urloff)
		c = f.read (1)
		i = 0
		url = ""

		while (c[0] != '\0' and currecoff + urloff + i + 1 < filesize):
			url += c[0]
			f.seek (currecoff + urloff + i + 1)
			c = f.read (1)
			i += 1

		h.url = url

		newoffset = 0
		filename = ""

		if (fileversion >= 5):
			newoffset = currecoff + 0x3C
		else:
			newoffset = currecoff + 0x40

		i = 0
		f.seek (newoffset)
		tmp = f.read (4)
		filenameoff = self.bah_to_i (tmp)
		filenameoff += currecoff

		if (filenameoff > currecoff + 0x3C):
			f.seek (filenameoff)
			c = f.read (1)

			while (c[0] != '\0' and filenameoff + i + 1 < filesize):
				filename += c[0]
				f.seek (filenameoff + i + 1)
				c = f.read (1)
				i += 1

			h.filename = filename

		newoffset = 0

		if (fileversion >= 5.2):
			newoffset = currecoff + 0x38
		elif (fileversion >= 5):
			newoffset = currecoff + 0x39
		else:
			newoffset = currecoff + 0x3C

		f.seek (newoffset)
		c = f.read (1)

		dirname = ""
		dirnameoff = int (ord(c[0]))	
		
		if (0x50 + (12 * dirnameoff) + 8 < filesize):
			f.seek (0x50+(12 * dirnameoff))
			dirname = f.read (8)

		h.dirname = dirname

		newoffset = 0
		httpheaders = ""

		if (fileversion >= 5):
			newoffset = currecoff + 0x44
		else:
			newoffset = currecoff + 0x48

		f.seek (newoffset)
		tmp = f.read (4)
		httpheaderoff = self.bah_to_i (tmp) + currecoff

		if (httpheaderoff > currecoff + 0x44):
			f.seek (httpheaderoff)
			c = f.read (1)

			while (c[:1] != '\0' and httpheaderoff + i + 1 < currecoff + reclen and httpheaderoff + i + 1 < filesize):
				httpheaders += c[0]
				f.seek (httpheaderoff + i + 1)
				c = f.read (1)
				i += 1
		
			h.headers = httpheaders
		
		self.browserHistory.append (h)

	def parse (self):
		f = open (self.filename, "rt")

		# Get file version
		f.seek (0x18)
		version = f.read (4)
#		print "version: %s" %version

		# Get file size
		f.seek (0x1C)
		tmp = f.read (4);
		filesize = self.bah_to_i (tmp)	
#		print "filesize: %d" %filesize

		# Get hashoff
		f.seek (0x20)
		tmp = f.read (4);
		hashoff = self.bah_to_i (tmp)
#		print "hashoff: %d" %hashoff

		while (hashoff != 0):
			f.seek (hashoff + 8)
			tmp = f.read (4)
			nexthashoff = self.bah_to_i (tmp)
			
			f.seek (hashoff + 4)
			tmp = f.read (4)
			hashsize = self.bah_to_i (tmp)
			hashsize *= 0x80;

			offset = hashoff + 16
			while (offset < hashoff + hashsize):
				f.seek (offset)
				hashrecflags = f.read (4)

				f.seek (offset + 4)
				tmp = f.read (4)
				currecoff = self.bah_to_i (tmp)

				if (hashrecflags[:1] != 0x03 and currecoff != 0xBADF00D):
					f.seek (currecoff)
					tmp = f.read (4)

					if (tmp[:4] == 'REDR'):
						self.parseRedr (f, currecoff, filesize, tmp)
					elif (tmp[:3] == 'URL' or tmp[:4] == 'LEAK'):
						self.parseUrl (f, currecoff, filesize, tmp, version)

				offset += 8

			hashoff = nexthashoff

		f.close ()
